/* implements brachand bound algorithms */

#include "tsp.h"

struct node {
        unsigned int bound;
        unsigned int cur_weight;
        unsigned int best_estimate;
        unsigned int level;
        unsigned char *path;
};

void estimate_best(void) {
        /* fills in array where minimum cost of leaving every 
         * city is stored
	 */
        register int i,j;
        unsigned int tmp;

        best_array = (unsigned int*)calloc(cities, sizeof(unsigned int));
        printf("\n Initial estimated path is %ld\n",best_pathlength);

        for(i=0; i<cities; i++) {
                tmp = UINT_MAX;
                for(j=0; j<cities; j++) {
                        if(i==j)
                                continue;
                        if(i < j) {
                                if( tmp > main_edge_array[j][i] )
                                        tmp = main_edge_array[j][i];
                        }
                        else {
                                if( tmp > main_edge_array[i][j] )
                                        tmp = main_edge_array[i][j];
                        }
                }
                best_array[i] = tmp;
                best_possible += tmp;
        }
        printf("absolute minimum is %u\n",best_possible);
}

int bb(void) {
        register int i,j,n;
        unsigned int tmp;
        unsigned int tmp_weight,tmp_bound;
        struct node **queue, *work_node, *child;
        /* **queue will be heap array root pointer, I will use */
        /* queue[0] to store current number of elements in queue */

        if(!(full_best_path = (char*)malloc(cities*sizeof(char))))
                handle_error("memory allocation failed");
        estimate_best();
        if(!(queue = (struct node**)calloc(HEAP_SIZE,sizeof(struct node*))))
                        handle_error("memory allocation failed");
        queue[0] = (struct node*)calloc(1,sizeof(struct node));
        /* queue[0] is used for various heap internal info */
        work_node = (struct node*)calloc(1,sizeof(struct node));
        work_node->path = (char*)calloc(cities, sizeof(char));
        work_node->level = work_node->cur_weight =  0;
        work_node->best_estimate = best_array[0];
        work_node->bound = best_possible;
        queue_insert(queue,work_node);

        while((work_node = (struct node*)queue_remove(queue))) {
                /* remove first member from priority queue */

                if( work_node->bound < best_pathlength ) {
                        if( queue[0]->best_estimate > HEAP_SIZE -cities) {
                                /* runing out of memory, let's calculate this branch */
                                /* to an end with depth first algorithm              */
#ifdef INFO
                                printf("Out of memory !! Switching to depth first\n");
#endif
                                bb_dive(work_node);
                        }

                        else {
                                for(i=1; i<cities; i++) {
                                        for(n=0; n<work_node->level+1; n++) {
                                                /* exclude nodes that are already in path */
                                                if( work_node->path[n] == i )
                                                        goto END;
                                        }
                                        /* calculate bound value */
                                        if( work_node->path[work_node->level] < i )
                                                tmp_weight = main_edge_array[i][work_node->path[work_node->level]];
                                        else
                                                tmp_weight = main_edge_array[work_node->path[work_node->level]][i];
                                        tmp_weight += work_node->cur_weight;
                                        tmp_bound = best_possible - work_node->best_estimate;
                                        tmp_bound += tmp_weight;
                                        /* add node to queue if bound value is suitable */
                                        if( tmp_bound < best_pathlength ) {
                                                if( work_node->level == cities - 2 ) {
                                                        tmp = tmp_bound + main_edge_array[i][0];
                                                        tmp -= best_array[i];
                                                        if( tmp < best_pathlength ) {
                                                                printf("new best : %u\n",tmp);
                                                                work_node->path[cities-1] = i;
                                                                for(j=0; j<cities; j++) {
                                                                        full_best_path[j]=work_node->path[j];
                                                                        printf("%u->",full_best_path[j]);
                                                                }
                                                                printf("\n");
                                                                best_pathlength = tmp;
                                                        }
                                                        continue;
                                                }
                                                /* create , initialize and add new member to queue */
                                                if(!(child = (struct node*)calloc(1,sizeof(struct node))))
                                                        handle_error("memory allocation failed");
                                                if(!(child->path = (char*)calloc(work_node->level+2,sizeof(char))))
                                                        handle_error("memory allocation failed");;
                                                child->level = work_node->level + 1;
                                                child->cur_weight = tmp_weight;
                                                child->bound = tmp_bound;
                                                child->best_estimate = work_node->best_estimate+best_array[i];
                                                memcpy(child->path, work_node->path, work_node->level+1);
                                                child->path[child->level] = i;
                                                queue_insert(queue,child);
                                        }
                                        END:;
                                }
                        }
                }

                free(work_node->path);
                free(work_node);
        }

	return(0);
}

void bb2_master(void) {
        if(!(cur_path = (char*)calloc(cities,sizeof(char))))
                handle_error("memory allocation failed");
        if(!(full_best_path = (char*)malloc(cities*sizeof(char))))
                handle_error("memory allocation failed");
        if(!(used = (char*)calloc(cities,sizeof(char))))
                handle_error("memory allocation failed");

        estimate_best();
        cur_length = 0;
        used[0] = 1;
        bb_complete(1);
}


void bb_dive(struct node *branch) {
        register int i;
        unsigned int tmp_best;

        if(!(cur_path = (char*)malloc(cities*sizeof(char))))
                handle_error("memory allocation failed");
        if(!(used = (char*)calloc(cities,sizeof(char))))
                handle_error("memory allocation failed");

        cur_length = 0;
        tmp_best = best_possible;
        best_possible -= branch->best_estimate;
        cur_length = branch->cur_weight;
        for(i=0;i<=branch->level;i++) {
                used[branch->path[i]] = 1;
                cur_path[i] = branch->path[i];
        }

        bb_complete(branch->level+1);
        best_possible = tmp_best;

        free(cur_path);
        free(used);
}

void bb_complete(int done) {
        register int i;
        unsigned int tmp_length,tmp_best;


        if( done == cities ) {
                /* add weight of going back to starting point */
                cur_length += main_edge_array[cur_path[cities-1]][0];
                if( cur_length < best_pathlength ) {
                        printf("new best : %ld\n",cur_length);
                        for(i=0;i<cities;i++) {
                                full_best_path[i] = cur_path[i];
                                printf("%d->",cur_path[i]);
                        }
                        printf("\n");
                        fflush(stdout);
                        best_pathlength = cur_length;
                }
        }

        else  {
                tmp_best = best_possible;
                tmp_length = cur_length;
                for(i=1; i<cities; i++) {
                        if( used[i] == 0 ) {
                                best_possible = tmp_best;
                                cur_length = tmp_length;
                                if( i > cur_path[done-1] )
                                        cur_length += main_edge_array[i][cur_path[done-1]];
                                else
                                        cur_length += main_edge_array[cur_path[done-1]][i];
                                best_possible -= best_array[i];
                                if( (cur_length + best_possible) < best_pathlength ) {
                                        used[i] = 1;
                                        cur_path[done] = i;
                                        bb_complete(done+1);
                                        used[i] = 0;
                                }
                        }
                }
        }
}

struct node *queue_insert(struct node **queue, struct node *new) {
        register int i;

        i = ++queue[0]->best_estimate;
        /* we insert new node to the end of the array */
        while( i > 1 && queue[i>>1]->bound > new->bound ) {
                if( queue[i>>1]->bound == new->bound && queue[i>>1]->level > new->level )
                        break;
                queue[i] = queue[i>>1];
                i>>=1;
        }
        queue[i] = new;
	return new;
}

struct node *queue_remove(struct node **queue) {
        register int i,j=0;

        struct node *tmp_min,*tmp_last;

        if( queue[0]->best_estimate == 0 )
                return(NULL);

        tmp_min = queue[1];
        tmp_last = queue[queue[0]->best_estimate--];

        for(i=1; (i<<1)<=queue[0]->best_estimate; i=j ) {
                j = i<<1;
                if( j != queue[0]->best_estimate ) {
                        if( queue[j+1]->bound < queue[j]->bound )
                                j++;
                        else if( queue[j+1]->bound == queue[j]->bound && queue[j+1]->level > queue[j]->level )
                                j++;
                }
                if( tmp_last->bound > queue[j]->bound )
                        queue[i] = queue[j];
                else
                        break;
        }
        *(queue+i) = tmp_last;
        return(tmp_min);
}

