#include "tsp.h"

struct unused_vertix {
        unsigned int vertix;
        struct unused_vertix *next;
};

struct max_node {
        struct unused_vertix *next;
        struct unused_vertix *prev;
        unsigned int city1;
        unsigned int city2;
        long int weight;
};

struct unused_vertix* init_unused_vertix_list( int starting_point ) {
        register int i;
        struct unused_vertix *tmp,*v_list;
        /* this function will generate linked list, containing all          
         * the unused vertixes, here it means all the vertixes but starting 
         * point                                                            
	 */

        if(!(v_list = (struct unused_vertix*)malloc(sizeof(struct unused_vertix))))
                handle_error("memory allocation failed in init_vertix_list");
        tmp = v_list;

        for(i=0; i<cities; i++) {
                if( i==starting_point )
                        continue;
                v_list->next = (struct unused_vertix*)malloc(sizeof(struct unused_vertix));
                v_list->vertix = i;
                if( i!=(cities-1) )
                        v_list = v_list->next;
                else
                        v_list->next=NULL;
        }

	return(tmp);
}

int greedy(int starting_point) {
        /* This is basically implementation of Dijkstra's greedy 
         * algorithm, with possibility to try multiple starting  
         * points                                                
	 */
        struct unused_vertix *unused_list,*base_node;
        /* unused_list and base_node are pointers to linked list 
         * where I keep vertixes that are unused                 
	 */
        register struct unused_vertix *tmp_node;
        unsigned int path_index[2];
        /* path_index is used to keep track how many members are in 
         * answer_array[n][0] & [n][1], from those arrays we will   
         * generate path                                            
	 */
        register int i;
        struct max_node *cur_max;
        /* structure were I keep info about best edge found yet 
         * it also contains pointers to unused edges list, that are needed 
         * for removing node                                               
	 */
        unsigned int tmp;
        unsigned long int pathlen=0;
        register int array_counter = 0;
        unsigned int work[2];
        /* our working stack, there can be max. 2 cities at work once */
        unsigned int **answer_array;
        /* answer_array[n][0] = cities that come from work[0] stack 
         * answer_array[n][1] = cities that come from work[1] stack 
         * answer_array[n][2] = recalculated path lengths from starting point 
	 */

        answer_array = (unsigned int**)calloc(cities,sizeof(unsigned int*));
        for(i=0;i<cities;i++)
                *(answer_array + i) = (unsigned int*)calloc(cities,sizeof(unsigned int));
        cur_max = (struct max_node*)malloc(sizeof(struct max_node));

        work[0] = work[1] = starting_point; /* initialize */
        path_index [0] = path_index[1] = 0;
        cur_max->prev = NULL;
        cur_max->weight = UINT_MAX;

        /* we have to copy main_e_array[starting_point][N] somewhere     
         * else so we don't have to change values in main data matrix    
         * and therefore make trying multiple starting points imposible  
	 */
        for(i=0; i<cities; i++) {
                if( i < starting_point )
                       answer_array[i][2] = main_edge_array[starting_point][i];
                else
                        answer_array[i][2] = main_edge_array[i][starting_point];
        }

        /* initialize unused vertix list */
        base_node = init_unused_vertix_list(starting_point);
        unused_list = base_node;

        while( base_node!=NULL ){ /* there are no unused nodes left => we are done */
             for(i=0; i<2; i++) {
                /* we have 2 cities in stack at any given time, and we will compare 
                 * all the paths from both cities in stack to any vertix in unused  
                 * vertix list + path length to starting point                      
		 */
                unused_list = base_node;
                tmp_node = NULL;

                do {    /* go through unused vertix list 
                         * tmp_node is used to keep pointer to previous node 
			 */
                        if( work[i] > unused_list->vertix )  /* hack to get addressing right */
                                tmp = answer_array[work[i]][2] + main_edge_array[work[i]][unused_list->vertix];
                        else
                                tmp = answer_array[work[i]][2] + main_edge_array[unused_list->vertix][work[i]];
                        /* we will calculate weight of every possible node relative to starting 
                         * point 
			 */

                        if( tmp < cur_max->weight ) {
                                /* we have found new best, lets update current max */
                                cur_max->next = unused_list->next;
                                cur_max->prev = tmp_node;
                                cur_max->weight = tmp;
                                cur_max->city1 = i;
                                cur_max->city2 = unused_list->vertix;
                        }
                        tmp_node = unused_list;
                        unused_list = unused_list->next;
                } while( unused_list != NULL );
             }
             /* at this point we have new shortest node */
             answer_array[path_index[cur_max->city1]++][cur_max->city1] = work[cur_max->city1];

             if( work[cur_max->city1] > cur_max->city2 )
                pathlen += main_edge_array[work[cur_max->city1]][cur_max->city2];
             else
		pathlen += main_edge_array[cur_max->city2][work[cur_max->city1]];
             answer_array[cur_max->city2][2] = cur_max->weight;
                /* update weight of edge from starting point to newest shortest path 
                 * in data array 
		 */
             if( cur_max->prev == NULL ) { /* cutting first node from list */
                free( base_node );
                base_node = cur_max->next;
                }
             else {
                free( ((struct unused_vertix*)cur_max->prev)->next );
                ((struct unused_vertix*)cur_max->prev)->next = cur_max->next;
             }
                /* remove used node */
             work[cur_max->city1] = cur_max->city2;
                /* put new city to working stack */
             cur_max->weight = UINT_MAX;
             array_counter++;
             unused_list = base_node;
        }
        /* main cycle ends, at this point we should have all the edges, but the last one 
         * which is waiting for us in our working stack                                  
	 */

        answer_array[path_index[0]][0] = work[0];
        answer_array[path_index[1]][1] = work[1];
        if( work[0] > work[1] )
                pathlen += main_edge_array[work[0]][work[1]];
        else
                pathlen += main_edge_array[work[1]][work[0]];

        if( best_path == NULL ) {
                /* this is true if we don't want to try multiple  
                 * starting points, so answer is written straight 
                 * out                                            
		 */
                for(i=0;i<=path_index[0];i++)
                        printf(" %u ->",answer_array[i][0]);
                for(i=path_index[1];i>=0;--i)
                        printf(" %u ->",answer_array[i][1]);
                printf("\n Total length of path : %ld\n",pathlen);
        }

        else if( (pathlen < best_pathlength) || (best_pathlength == 0)) {
                /* this is for searching shortest path with multiple 
                 * starting points                                   
		 */
                best_pathlength = pathlen;
                printf(" found new best -> %ld\n ",pathlen);
                for(i=0;i<=path_index[0];i++)
                        best_path[i] = answer_array[i][0];
                for(i=path_index[1];i>=0;--i)
                        best_path[cities-i] = answer_array[i][1];
        }

        for(i=0; i<cities; i++)
                free(*(answer_array+i));
        free(answer_array);
        free(cur_max);
	return(0);
}



